#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
import json
import logging

from container_app_conf.source import FilesystemSource

LOGGER = logging.getLogger(__name__)


class JsonSource(FilesystemSource):
    """
    Data source utilizing YAML files
    """
    DEFAULT_FILE_EXTENSIONS = ['json']

    def _load(self, file_path: str) -> dict:
        with open(file_path, 'r') as file:
            return json.load(file)

    def _write_reference(self, reference: dict, file_path: str):
        with open(file_path, "w") as file:
            json.dump(reference, file, indent=2)



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
from container_app_conf import ConfigBase
from container_app_conf.entry.bool import BoolConfigEntry
from tests import TestBase, TestConfigBase


class TestConfigBase2(ConfigBase):
    BOOL = BoolConfigEntry(
        key_path=["test", "bool"],
        # default=False,
        example=True
    )


class TestSingleton(TestBase):

    def test_singleton(self):
        assert not TestConfigBase() == TestConfigBase2()
        assert TestConfigBase() == TestConfigBase()
        assert TestConfigBase2() == TestConfigBase2()

    def test_singleton_config_entry(self):
        conf1 = TestConfigBase()
        conf2 = TestConfigBase()

        conf1.INT.value = 1
        conf2.INT.value = 2

        self.assertEqual(conf1.INT.value, conf2.INT.value)

    def test_instance_config_entry(self):
        conf1 = TestConfigBase()
        conf2 = TestConfigBase(singleton=False)
        conf3 = TestConfigBase(singleton=False)

        conf1.INT.value = 1
        conf2.INT.value = 2
        conf3.INT.value = 3

        self.assertNotEqual(conf1.INT.value, conf2.INT.value)
        self.assertNotEqual(conf1.INT.value, conf3.INT.value)
        self.assertNotEqual(conf2.INT.value, conf3.INT.value)



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

from pathlib import Path

from py_range_parse import Range

from container_app_conf.entry.bool import BoolConfigEntry
from container_app_conf.entry.date import DateConfigEntry
from container_app_conf.entry.file import FileConfigEntry, DirectoryConfigEntry
from container_app_conf.entry.float import FloatConfigEntry
from container_app_conf.entry.int import IntConfigEntry
from container_app_conf.entry.range import RangeConfigEntry
from container_app_conf.entry.timedelta import TimeDeltaConfigEntry
from tests import EntryTestBase


class EntryTest(EntryTestBase):

    def test_bool_entry(self):
        config_entry = BoolConfigEntry(key_path=["bool"])

        true_values = ["y", "yes", "true", "t", 1, True]
        false_values = ["n", "no", "false", "f", 0, False]
        invalid_values = ["hello", 2, 0.1]

        input_output = []

        for tv in true_values:
            input_output.append((tv, True))
        for fv in false_values:
            input_output.append((fv, False))
        for iv in invalid_values:
            input_output.append((iv, ValueError))

        self.assert_input_output(config_entry, input_output)

    def test_int_entry(self):
        config_entry = IntConfigEntry(key_path=["int"])
        input_output = [
            ("5", 5),
            (5, 5),
            ("-3", -3),
            (-3, -3)
        ]

        self.assert_input_output(config_entry, input_output)

    def test_float_entry(self):
        config_entry = FloatConfigEntry(key_path=["float"])
        input_output = [
            ("5", 5.0),
            (5, 5.0),
            ("-3.0", -3.0),
            (-3.0, -3.0),
            (1.2, 1.2),
            ("1.2", 1.2),
            ("3%", 0.03)
        ]

        self.assert_input_output(config_entry, input_output)

    def test_date_entry(self):
        from datetime import datetime
        from dateutil.tz import tzutc

        config_entry = DateConfigEntry(key_path=["date"])
        input_output = [
            ("2008-09-03T20:56:35.450686Z", datetime(2008, 9, 3, 20, 56, 35, 450686, tzinfo=tzutc())),
            ("2008-09-03", datetime(2008, 9, 3, 0, 0, 0, 0)),
        ]

        self.assert_input_output(config_entry, input_output)

    def test_timedelta_entry(self):
        from datetime import timedelta

        config_entry = TimeDeltaConfigEntry(key_path=["timedelta"])
        input_output = [
            ("20:56:35", timedelta(hours=20, minutes=56, seconds=35)),
            ("32m", timedelta(minutes=32)),
            ("4h0m3s", timedelta(hours=4, minutes=0, seconds=3)),
            ("4h3s", timedelta(hours=4, minutes=0, seconds=3)),
            ("4:13", timedelta(hours=0, minutes=4, seconds=13)),
        ]

        self.assert_input_output(config_entry, input_output)

    def test_file_entry(self):
        config_entry = FileConfigEntry(key_path=["file"])
        input_output = [
            ("/tmp/test", Path("/tmp/test")),
            ("./test", Path("./test")),
            ("/something/", AssertionError),
        ]

        self.assert_input_output(config_entry, input_output)

        config_entry = FileConfigEntry(key_path=["file"],
                                       check_existence=True)
        input_output = [
            ("/tmp/test", FileNotFoundError),
        ]

        self.assert_input_output(config_entry, input_output)

    def test_directory_entry(self):
        config_entry = DirectoryConfigEntry(key_path=["directory"])
        input_output = [
            ("/tmp", AssertionError),
            ("/tmp/", Path("/tmp")),
            ("./test/", Path("./test")),
            ("/something/", Path("/something")),
        ]

        self.assert_input_output(config_entry, input_output)

        config_entry = DirectoryConfigEntry(key_path=["directory"],
                                            check_existence=True)

        input_output = [
            ("./", Path("./")),
        ]

        self.assert_input_output(config_entry, input_output)

    def test_range_entry(self):
        config_entry = RangeConfigEntry(key_path=["range"])
        input_output = [
            ("[-5..5]", Range(-5, 5)),
        ]

        self.assert_input_output(config_entry, input_output)



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
from typing import List

from container_app_conf import ConfigEntry


def find_duplicates(l: list) -> []:
    """
    Finds duplicate entries in the given list
    :param l: the list to check
    :return: map of (value -> list of indexes)
    """
    if not len(l) != len(set(l)):
        return []

    # remember indexes of items with equal hash
    tmp = {}
    for i, v in enumerate(l):
        if v in tmp.keys():
            tmp[v].append(i)
        else:
            tmp[v] = [i]

    result = {}
    for k, v in tmp.items():
        if len(v) > 1:
            result[k] = v

    return result


def generate_reference_config(config_entries: List[ConfigEntry]) -> {}:
    """
    Generates a dictionary containing the expected config tree filled with default and example values
    :return: a dictionary containing the expected config tree
    """
    config_tree = {}
    for entry in config_entries:
        current_level = config_tree
        for path in entry.key_path[:-1]:
            if path not in current_level:
                current_level[path] = {}
            current_level = current_level[path]

        current_level[entry.key_path[-1]] = entry._type_to_value(entry.example)

    return config_tree



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
from datetime import datetime
from datetime import timedelta
from pathlib import Path

from dateutil.tz import tzutc

from container_app_conf.entry.date import DateConfigEntry
from container_app_conf.entry.file import FileConfigEntry
from container_app_conf.entry.float import FloatConfigEntry
from container_app_conf.entry.int import IntConfigEntry
from container_app_conf.entry.list import ListConfigEntry
from container_app_conf.entry.timedelta import TimeDeltaConfigEntry
from tests import EntryTestBase


class ListEntryTest(EntryTestBase):

    def test_str_list_entry_custom_delimiter(self):
        config_entry = ListConfigEntry(item_type=IntConfigEntry,
                                       key_path=["int_list"],
                                       delimiter="::")
        input_output = [
            ("1::2::3", [1, 2, 3])
        ]

        self.assert_input_output(config_entry, input_output)

    def test_int_list_entry(self):
        config_entry = ListConfigEntry(item_type=IntConfigEntry,
                                       key_path=["int_list"])
        input_output = [
            ("", []),
            ("5", [5]),
            ("1,2,3", [1, 2, 3])
        ]

        self.assert_input_output(config_entry, input_output)

    def test_float_list_entry(self):
        config_entry = ListConfigEntry(item_type=FloatConfigEntry,
                                       key_path=["float_list"])
        input_output = [
            ("", []),
            ("5", [5.0]),
            ("", []),
            ("1,2.5,3", [1.0, 2.5, 3.0])
        ]

        self.assert_input_output(config_entry, input_output)

    def test_date_list_entry(self):
        config_entry = ListConfigEntry(item_type=DateConfigEntry,
                                       key_path=["date_list"])

        input_example_1 = ["2008-09-03T20:56:35.450686Z", "2008-09-03"]
        output_example_1 = [datetime(2008, 9, 3, 20, 56, 35, 450686, tzinfo=tzutc()),
                            datetime(2008, 9, 3, 0, 0, 0, 0)]
        input_output = [
            (None, None),
            (input_example_1, output_example_1),
            (",".join(input_example_1), output_example_1)
        ]

        self.assert_input_output(config_entry, input_output)

    def test_timedelta_entry(self):
        config_entry = ListConfigEntry(item_type=TimeDeltaConfigEntry,
                                       key_path=["timedelta_list"])
        input_output = [
            (None, None),
            (
                ["20:56:35", "32m", "4h0m3s", "4h3s", "4:13"],
                [timedelta(hours=20, minutes=56, seconds=35),
                 timedelta(minutes=32),
                 timedelta(hours=4, minutes=0, seconds=3),
                 timedelta(hours=4, minutes=0, seconds=3),
                 timedelta(hours=0, minutes=4, seconds=13)]
            )
        ]

        self.assert_input_output(config_entry, input_output)

    def test_file_entry(self):
        config_entry = ListConfigEntry(item_type=FileConfigEntry,
                                       item_args={
                                           "check_existence": False
                                       },
                                       key_path=["file_list"])
        example1 = "/tmp/file"
        example2 = "./file"
        example3 = [example1, example2]
        input_output = [
            (None, None),
            ("/tmp/", AssertionError),
            (example1, [Path(example1)]),
            (example2, [Path(example2)]),
            (",".join(example3), [Path(example1), Path(example2)]),
        ]

        self.assert_input_output(config_entry, input_output)



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
import copy
import logging
from typing import Dict, List

from container_app_conf.const import DEFAULT_CONFIG_FILE_PATHS
from container_app_conf.entry import ConfigEntry
from container_app_conf.source import DataSource
from container_app_conf.util import find_duplicates, generate_reference_config

LOGGER = logging.getLogger(__name__)


class ConfigBase:
    """
    Config base class.
    Extend this in your application.
    """

    _instances = {}

    def __new__(cls, data_sources: List[DataSource] = None,
                validate: bool = True,
                singleton: bool = True,
                write_reference: bool = True):
        """
        Creates a config object and reads configuration.
        :param data_sources: list of data sources to use. The first value that holds a value for a specific
                             config entry overshadows other data sources.
        :param validate: if validation should be run (can be disabled for tests)
        :param singleton: if the returned instance should be a singleton
        :param write_reference: Whether to write a reference configuration for all used data sources
        """
        if singleton:
            if cls._instances.get(cls, None) is None:
                instance = super(ConfigBase, cls).__new__(cls)
                cls._instances[cls] = instance
            else:
                instance = cls._instances[cls]
        else:
            instance = super(ConfigBase, cls).__new__(cls)

        self = instance
        self._config_entries = self._find_config_entries()

        if not singleton:
            # copy class attribute to instance to overshadow class attributes
            instance_attributes = {}
            for name, attribute in self._config_entries.items():
                attribute_copy = copy.deepcopy(attribute)
                self.__dict__.setdefault(name, attribute_copy)
                key = "_".join(attribute.key_path)
                instance_attributes[key] = attribute_copy
            # update config_entries list to reflect instance attributes
            self._config_entries = instance_attributes

        if data_sources is None:
            # set default data sources
            from container_app_conf.source.env_source import EnvSource
            from container_app_conf.source.yaml_source import YamlSource

            self.data_sources = [
                EnvSource(),
                YamlSource(cls.__name__)
            ]
        else:
            self.data_sources = data_sources

        if write_reference:
            reference_config = generate_reference_config(list(self._config_entries.values()))
            for data_source in self.data_sources:
                data_source.write_reference(reference_config)

        self.load_config(validate)

        return instance

    def load_config(self, validate: bool):
        """
        Loads the configuration from all available sources
        """
        for source in reversed(self.data_sources):
            for entry in self._config_entries.values():
                if source.has(entry):
                    entry.value = source.get(entry)

        # reset all entries to make sure None constraints are fulfilled
        for entry in self._config_entries.values():
            entry.value = entry.value

        if validate:
            self.validate()

    def validate(self):
        """
        Validates the current configuration and throws an exception if something is wrong
        """
        return

    def _find_config_entries(self) -> Dict[str, ConfigEntry]:
        """
        Detects config entry constants in this class
        :return: list of config entries
        """
        entries = {}
        for name, attribute in dict(map(lambda x: (x, getattr(self, x)), dir(self))).items():
            if isinstance(attribute, ConfigEntry):
                entries[name] = attribute

        entry_env_keys = list(map(lambda x: "->".join(x.key_path), entries.values()))
        duplicates = find_duplicates(entry_env_keys)
        if len(duplicates) > 0:
            clashing = ", ".join(duplicates.keys())
            raise ValueError("Key paths must be unique! Clashing paths: {}".format(clashing))

        return entries
